const fetch = require('node-fetch');
const ndjson = require('ndjson');
const serialize = require('../../metadata/serialize');

const DEFAULT_REQUEST_TIMEOUT = 15000;

class NDJsonResponse {
  constructor(fetchResponse) {
    this.fetchResponse = fetchResponse;
  }

  stream() {
    return this.fetchResponse.body.pipe(ndjson.parse());
  }

  values() {
    return new Promise((resolve, reject) => {
      const vals = [];
      const stream = this.stream();
      stream.on('data', val => {
        vals.push(val);
      });
      stream.on('error', reject);
      stream.on('finish', () => {
        resolve(vals);
      });
    });
  }
}

class RestError extends Error {
  constructor(response, responseBody) {
    super(response.statusText + '\n' + responseBody);
    this.statusCode = response.status;
    this.response = response;
  }
}

class RestClient {

  constructor(options) {
    this.apiUrl = options.apiUrl || '';
    this.requestTimeout = options.requestTimeout != null ? options.requestTimeout : DEFAULT_REQUEST_TIMEOUT;
  }

  _makeUrl(path) {
    const absPath = path.startsWith('/') ? path : '/' + path;
    return this.apiUrl + absPath;
  }

  req(path, args = {}) {
    const fullUrl = this._makeUrl(path);
    args = Object.assign({ timeout: this.requestTimeout }, args);
    return fetch(fullUrl, args).then(response => {
      if (!response.ok) {
        return response.text().then(responseBody => {
          throw new RestError(response, responseBody);
        });
      }
      return response;
    });
  }

  getRequest(path) {
    return this.req(path);
  }

  postRequest(path, body, isJSON = true) {
    return this.req(path, {
      method: 'POST',
      headers: isJSON ? { 'Content-Type': 'application/json' } : {},
      body: isJSON ? JSON.stringify(body) : body
    });
  }

  id(peerId) {
    let path = 'id';
    if (peerId != null) {
      path += '/' + peerId;
    }
    return this.getRequest(path).then(r => r.json());
  }

  ping(peerId) {
    return this.getRequest(`ping/${ peerId }`).then(response => true);
  }

  publish(opts, ...statements) {
    const { namespace, compound } = opts;
    const statementNDJSON = statements.map(s => JSON.stringify(s)).join('\n');

    let path = `publish/${ namespace }`;
    if (compound != null) {
      path += '/' + compound.toString();
    }

    return this.postRequest(path, statementNDJSON, false).then(r => r.text()).then(text => text.split('\n').filter(text => text.length > 0));
  }

  statement(statementId) {
    return this.getRequest(`stmt/${ statementId }`).then(r => r.json());
  }

  query(queryString, remotePeer) {
    return this.queryStream(queryString, remotePeer).then(r => r.values());
  }

  queryStream(queryString, remotePeer) {
    let path = 'query';
    if (remotePeer != null) {
      path += '/' + remotePeer;
    }
    return this.postRequest(path, queryString, false).then(r => new NDJsonResponse(r));
  }

  merge(queryString, remotePeer) {
    return this.postRequest(`merge/${ remotePeer }`, queryString, false).then(r => r.text()).then(resp => {
      const counts = resp.split('\n').filter(line => line.length > 0).map(line => Number.parseInt(line));
      const [statementCount, objectCount] = counts;
      return { statementCount, objectCount };
    });
  }

  push(queryString, remotePeer) {
    return this.postRequest(`push/${ remotePeer }`, queryString, false).then(r => r.text()).then(resp => {
      const counts = resp.split('\n').filter(line => line.length > 0).map(line => Number.parseInt(line));
      const [statementCount, objectCount] = counts;
      return { statementCount, objectCount };
    });
  }

  delete(queryString) {
    return this.postRequest('delete', queryString, false).then(r => r.text()).then(Number.parseInt);
  }

  putData(...objects) {
    const body = objects.filter(o => o != null).map(o => {
      if (o instanceof Buffer) return o;
      try {
        return serialize.encode(o);
      } catch (err) {
        console.error('Serialization error: ' + err.message);
        console.error('Failed Object: ' + JSON.stringify(o, null, 2));
        throw err;
      }
    }).filter(buf => buf.length > 0).map(buf => ({ data: buf.toString('base64') })).map(obj => JSON.stringify(obj)).join('\n');

    return this.postRequest('data/put', body, false).then(r => r.text()).then(response => response.split('\n').filter(str => str.length > 0));
  }

  getData(objectId) {
    return this.getRequest(`data/get/${ objectId }`).then(r => r.json()).then(o => Buffer.from(o.data, 'base64')).then(bytes => {
      try {
        return serialize.decode(bytes);
      } catch (err) {
        return bytes;
      }
    });
  }

  listPeers() {
    return this.getRequest('dir/list').then(r => r.text()).then(s => s.split('\n').filter(line => line.length > 0));
  }

  getAuthorizations() {
    return this.getRequest('auth').then(r => r.json());
  }

  authorize(peerId, namespaces) {
    return this.postRequest(`auth/${ peerId }`, namespaces.join(','), false).then(r => r.text()).then(s => s.trim() === 'OK');
  }

  revokeAuthorization(peerId) {
    return this.authorize(peerId, []);
  }

  getStatus() {
    return this.getRequest('status').then(r => r.text()).then(validateStatus);
  }

  setStatus(status) {
    return this.postRequest(`status/${ status }`, '', false).then(r => r.text()).then(validateStatus);
  }

  getDirectoryId() {
    return this.getRequest('config/dir').then(r => r.text());
  }

  setDirectoryId(id) {
    return this.postRequest('config/dir', id, false).then(() => true);
  }

  getInfo() {
    return this.getRequest('config/info').then(r => r.text()).then(r => r.trim());
  }

  setInfo(info) {
    return this.postRequest('config/info', info, false).then(r => r.text());
  }

  getNATConfig() {
    return this.getRequest('config/nat').then(r => r.text()).then(s => s.trim());
  }

  setNATConfig(config) {
    return this.postRequest('config/nat', config, false).then(r => r.text()).then(s => s.trim() === 'OK');
  }

  getNetAddresses() {
    return this.getRequest('net/addr').then(r => r.text()).then(s => s.split('\n')).then(addrs => addrs.filter(s => s.length > 0));
  }

  shutdown() {
    return this.postRequest('shutdown', '', false).then(() => true).catch(err => {
      if (err.errno === 'ECONNRESET') {
        return true; // shutting down the node kills the request socket :)
      }
      throw err;
    });
  }
}

function validateStatus(status) {
  status = status.trim();
  switch (status) {
    case 'online':
    case 'offline':
    case 'public':
      return status;
  }
  throw new Error(`Unknown status: "${ status }"`);
}

module.exports = RestClient;