const P2PNode = require('./libp2p_node');
const PeerId = require('peer-id');
const PeerInfo = require('peer-info');
const Multiaddr = require('multiaddr');
const Multihash = require('multihashes');
const pb = require('../protobuf');
const pull = require('pull-stream');
const paramap = require('pull-paramap');
const { DEFAULT_LISTEN_ADDR, PROTOCOLS } = require('./constants');
const { inflateMultiaddr } = require('./identity');
const {
  protoStreamEncode,
  protoStreamDecode,
  peerInfoProtoMarshal,
  lookupResponseToPeerInfo,
  pullToPromise,
  pullRepeatedly,
  resultStreamThrough,
  objectIdsForQueryResult,
  expandQueryResult
} = require('./util');

const DEFAULT_INFO_MESSAGE = '(aleph)';

class MediachainNode {

  constructor(options) {
    let { peerId, dirInfo, listenAddresses } = options;
    if (listenAddresses == null) listenAddresses = [DEFAULT_LISTEN_ADDR];

    const peerInfo = new PeerInfo(peerId);
    listenAddresses.forEach(addr => {
      peerInfo.multiaddr.add(Multiaddr(addr));
    });

    this.infoMessage = options.infoMessage || DEFAULT_INFO_MESSAGE;

    this.p2p = new P2PNode({ peerInfo });
    this.directory = dirInfo;
    this.p2p.handle(PROTOCOLS.node.ping, this.pingHandler.bind(this));
    this.p2p.handle(PROTOCOLS.node.id, this.idHandler.bind(this));
  }

  start() {
    return this.p2p.start();
  }

  stop() {
    return this.p2p.stop();
  }

  get peerInfo() {
    return this.p2p.peerInfo;
  }

  setDirectory(dirInfo) {
    if (typeof dirInfo === 'string') {
      dirInfo = inflateMultiaddr(dirInfo);
    }
    this.directory = dirInfo;
  }

  setInfoMessage(message) {
    this.infoMessage = message;
  }

  register() {
    if (this.directory == null) {
      return Promise.reject(new Error('No known directory server, cannot register'));
    }

    const abortable = this.p2p.newAbortable();

    const req = {
      info: peerInfoProtoMarshal(this.p2p.peerInfo)
    };

    return this.p2p.dialByPeerInfo(this.directory, PROTOCOLS.dir.register).then(conn => {
      pull(pullRepeatedly(req, 5000 * 60), abortable, protoStreamEncode(pb.dir.RegisterPeer), conn, pull.onEnd(() => {
        console.log('registration connection ended');
      }));
      return true;
    });
  }

  lookup(peerId) {
    if (peerId instanceof PeerId) {
      peerId = peerId.toB58String();
    } else {
      // validate that string arguments are legit multihashes
      try {
        Multihash.fromB58String(peerId);
      } catch (err) {
        return Promise.reject(new Error(`Peer id is not a valid multihash: ${ err.message }`));
      }
    }

    // If we've already got an entry for this PeerId in our PeerBook,
    // because we already have a multiplex connection open to this peer,
    // use the existing entry.
    //
    // Note that when we close the peer multiplex connection, then entry is
    // automatically removed from the peer book, so we should never be returning
    // stale results.
    try {
      const peerInfo = this.p2p.peerBook.getByB58String(peerId);
      return Promise.resolve(peerInfo);
    } catch (err) {}

    if (this.directory == null) {
      return Promise.reject(new Error('No known directory server, cannot lookup'));
    }

    return this.p2p.dialByPeerInfo(this.directory, PROTOCOLS.dir.lookup).then(conn => pullToPromise(pull.values([{ id: peerId }]), protoStreamEncode(pb.dir.LookupPeerRequest), conn, protoStreamDecode(pb.dir.LookupPeerResponse), pull.map(lookupResponseToPeerInfo)));
  }

  _lookupIfNeeded(peer) {
    if (peer instanceof PeerInfo) {
      return Promise.resolve(peer);
    }
    if (typeof peer === 'string' && peer.startsWith('/')) {
      // try to decode as multiaddr. If it doesn't start with '/', it may be resolvable
      // as a multihash peer id via lookup()
      try {
        const peerInfo = inflateMultiaddr(peer);
        return Promise.resolve(peerInfo);
      } catch (err) {
        return Promise.reject(new Error(`Peer id is not a valid multiaddr: ${ err.message }`));
      }
    }

    return this.lookup(peer);
  }

  openConnection(peer, protocol) {
    return this._lookupIfNeeded(peer).then(maybePeer => {
      if (!maybePeer) throw new Error(`Unable to locate peer ${ peer }`);
      return maybePeer;
    }).then(peerInfo => this.p2p.dialByPeerInfo(peerInfo, protocol));
  }

  ping(peer) {
    return this.openConnection(peer, PROTOCOLS.node.ping).then(conn => pullToPromise(pull.values([{}]), protoStreamEncode(pb.node.Ping), conn, protoStreamDecode(pb.node.Pong), pull.map(_ => {
      return true;
    })));
  }

  pingHandler(protocol, conn) {
    pull(conn, protoStreamDecode(pb.node.Ping), protoStreamEncode(pb.node.Pong), conn);
  }

  idHandler(protocol, conn) {
    const response = {
      peer: this.peerInfo.id.toB58String(),
      info: this.infoMessage
    };

    pull(conn, protoStreamDecode(pb.node.NodeInfoRequest), pull.map(() => response), protoStreamEncode(pb.node.NodeInfo), conn);
  }

  remoteNodeInfo(peer) {
    return this.openConnection(peer, PROTOCOLS.node.id).then(conn => pullToPromise(pull.once({}), protoStreamEncode(pb.node.NodeInfoRequest), conn, protoStreamDecode(pb.node.NodeInfo)));
  }

  remoteQueryStream(peer, queryString) {
    return this.openConnection(peer, PROTOCOLS.node.query).then(conn => pull(pull.values([{ query: queryString }]), protoStreamEncode(pb.node.QueryRequest), conn, protoStreamDecode(pb.node.QueryResult), resultStreamThrough));
  }

  remoteQuery(peer, queryString) {
    return this.remoteQueryStream(peer, queryString).then(stream => new Promise((resolve, reject) => {
      pull(stream, pull.collect((err, results) => {
        if (err) return reject(err);
        resolve(results);
      }));
    }));
  }

  remoteData(peer, keys) {
    return this.remoteDataStream(peer, keys).then(stream => new Promise((resolve, reject) => {
      pull(stream, pull.collect((err, results) => {
        if (err) return reject(err);
        resolve(results);
      }));
    }));
  }

  remoteDataStream(peer, keys) {
    return this.openConnection(peer, PROTOCOLS.node.data).then(conn => pull(pull.once({ keys }), protoStreamEncode(pb.node.DataRequest), conn, protoStreamDecode(pb.node.DataResult), resultStreamThrough, pull.map(result => result.data)));
  }

  // local queries (NOT IMPLEMENTED -- NO LOCAL STORE)
  query(queryString) {
    throw new Error('Local statement db not implemented!');
  }

  data(keys) {
    throw new Error('Local datastore not implemented!');
  }

  remoteQueryWithDataStream(peer, queryString) {
    return this.remoteQueryStream(peer, queryString).then(resultStream => pull(resultStream, paramap((queryResult, cb) => {
      if (queryResult.value == null) {
        return cb(null, queryResult);
      }

      this._expandQueryResultData(peer, queryResult.value).then(result => cb(null, result)).catch(err => cb(err));
    })));
  }

  remoteQueryWithData(peer, queryString) {
    return this.remoteQueryWithDataStream(peer, queryString).then(stream => new Promise((resolve, reject) => {
      pull(stream, pull.collect((err, results) => {
        if (err) return reject(err);
        resolve(results);
      }));
    }));
  }

  _expandQueryResultData(peer, result) {
    const objectIds = objectIdsForQueryResult(result);
    if (objectIds.length < 1) return Promise.resolve(result);

    return this.remoteData(peer, objectIds).then(dataResults => expandQueryResult(result, dataResults));
  }
}

class RemoteNode {

  constructor(node, remotePeerInfo) {
    this.node = node;
    this.remotePeerInfo = remotePeerInfo;
  }

  ping() {
    return this.node.ping(this.remotePeerInfo);
  }

  query(queryString) {
    return this.node.remoteQuery(this.remotePeerInfo, queryString);
  }

  data(keys) {
    return this.node.remoteData(this.remotePeerInfo, keys);
  }

  queryWithData(queryString) {
    return this.node.remoteQueryWithData(this.remotePeerInfo, queryString);
  }
}

module.exports = {
  MediachainNode,
  RemoteNode
};